#include <iostream>
#include <fstream>
#include <unistd.h>          // usleep
#include <pthread.h>
#include <math.h>            // M_PI
#include <ntcan.h>
#include <iomanip>

#include <rtai_lxrt.h>
#include <sys/mman.h>

#include "Group.hh"
#include "globals.h"

#ifndef __CANBUS_H__
#define __CANBUS_H__

using namespace std;

#define NODE_MIN 1
#define NODE_MAX 15
#define NUM_NODES (NODE_MAX - NODE_MIN + 1)

class CANbus{
private:
  pthread_mutex_t busmutex;
  pthread_t canthread;
  pthread_mutex_t runmutex;
  int bus_run;

  int id;
  HANDLE handle;
  
  Group groups[NUM_GROUPS+1];
  Puck *pucks;
  int npucks;

  pthread_mutex_t trqmutex;
  pthread_mutex_t posmutex;
  long* trq;
  double* pos;

  int load(char* fn);
  int open();
  int check();
  
  int allow_message(int id, int mask);   
  int wake_puck(int id);
  
  int status(long* nodes);
  int parse(int32_t msgid, uint8_t* msg, int32_t msglen,
	    int32_t* nodeid, int32_t* property, long* value);
  int compile(int32_t property, long value, uint8_t *msg, int *msglen);
  
  int set_property(int32_t nid, int32_t property, long value, bool check);
  int get_property(int32_t nid, int32_t property, long* value);
  
  int read(int32_t* msgid, uint8_t* msg, int32_t* msglen, bool block);
  int send(int32_t  msgid, uint8_t* msg, int32_t  msglen, bool block);
  
  int send_torques();
  int read_positions();

public:

  CANbus();
  ~CANbus(){
    if(pucks!=NULL) delete pucks; 
    if(trq!=NULL) delete trq; 
    if(pos!=NULL) delete pos;
  }

  int init();
  int start();
  int stop();
  int run();
  void dump();
  void printpos();
  
  int send_torques(long* mtrq);
  int read_positions(double* mpos);
  int read_torques(long* mtrq);
  int send_positions(double* mpos);
  
  int limits(double jointVel, double tipVel, double elbowVel);
  friend void* canbus_handler(void* argv);
};

#endif
