#include "ControlLoop.hh"

static void* control_handler(void* argv){
  RT_TASK* task;
  RTIME samper, now;
  ControlLoop* ctrl_loop;

  ctrl_loop = (ControlLoop*)argv;

  rt_allow_nonroot_hrt();

  task = rt_task_init_schmod(nam2num("CTRLLP"), 0, 0, 0, SCHED_FIFO, 0xff);
  if(task == NULL){
    cerr << "control_handler: rt_task_init_schmod failed." << endl;
    return NULL;
  }
  
  if(mlockall(MCL_CURRENT | MCL_FUTURE) == -1){
    perror("control_handler: mlockall failed: ");
    return NULL;
  }

  rt_make_hard_real_time();

  rt_set_oneshot_mode();
  samper = nano2count( (RTIME)(ControlLoop::PERIOD*1000000000.0) );

  start_rt_timer(samper);
  now = rt_get_time();

  if(rt_task_make_periodic(task, now+samper, samper) != 0){
    cerr << "control_handler: rt_task_make_periodic failed." << endl;
    rt_task_delete(task);
    return NULL;
  }

  ctrl_loop->ctrl_fnc( ctrl_loop->ctrl_argv );

  stop_rt_timer();
  rt_task_delete(task);

  return argv;
}

ControlLoop::ControlLoop(){
  pthread_mutex_init(&mutex, NULL);
}

int ControlLoop::start(void* (*fnc)(void*), void* argv){

  ctrl_fnc = fnc;
  ctrl_argv = argv;

  lock();
  cls = CONTROLLOOP_RUN;
  unlock();

  if(pthread_create(&ctrlthread, NULL, control_handler, this)){
    cerr << "ControlLoop::start: pthread_create failed." << endl;
    lock();
    cls = CONTROLLOOP_STOP;
    unlock();
    return OW_FAILURE;
  }
  return OW_SUCCESS;
}

int ControlLoop::stop(){
  lock();
  cls = CONTROLLOOP_STOP;
  unlock();

  if(pthread_join(ctrlthread, NULL) != 0){
    cerr << "ControlLoop::stop: pthread_join failed." << endl;
    return OW_FAILURE;
  }
  return OW_SUCCESS;
}

int ControlLoop::state(){
  CONTROLLOOP_STATE s;
  lock();
  s = cls;
  unlock();
  return s;
}

void ControlLoop::wait(){rt_task_wait_period();}
