
#include "JointCtrl.hh"

#ifndef __JOINTCTRLPID_HH__
#define __JOINTCTRLPID_HH__

class JointCtrlPID : public JointCtrl{

private:
  double Kp, Kd, Ki;
  double laste;
  double se;         // Sum of error

public:

  JointCtrlPID(){reset();}
  
  double evaluate(double qs, double q, double dt){
    if(state() == Controller::RUN){
      set(qs);   // backup the command
      lock();
      double e = qs - q;
      double ed = (e - laste)/dt;
      se += e;
      laste = e;
      unlock();
      return Kp*e + Kd*ed + Ki*se;
    }
    return 0.0;
  }

  void reset(){       lock();  laste = 0;  unlock();  }

  istream& get(istream& s){    s >> Kp >> Kd >> Ki;    return s;  }
  
  ostream& put(ostream& s){
    s << "Kp: "  << setw(4) << Kp << "; "
      << "Kd: "  << setw(4) << Kd << "; "
      << "Ki: "  << setw(4) << Ki ;
    return s;
  }
};

#endif
