#include "DH.hh"

#include <openmath/R3.hh>
#include <openmath/SE3.hh>
#include <openmath/Inertia.hh>

#ifndef __LINK_HH__
#define __LINK_HH__

using namespace std;

class Link{
private:

  DH dh;
  R3 cog;
  double mass;
  Inertia inertia;

public:

  static const int L0 = 0;         // link 1 index
  static const int L1 = 1;         // link 1 index
  static const int L2 = 2;         // link 2 index
  static const int L3 = 3;         // link 3 index
  static const int L4 = 4;         // link 4 index
  static const int L5 = 5;         // link 5 index
  static const int L6 = 6;         // link 6 index
  static const int L7 = 7;         // link 7 index
#ifdef WRIST
  static const int Ln = Link::L7;
#else
  static const int Ln = Link::L4;
#endif

  Link(){}
  Link(const DH& dh, double m, const R3& cog, const Inertia& inertia){
    this->dh = dh;  this->mass = m;  this->inertia = inertia;  this->cog = cog;
  }

  operator SE3() const { return (SE3)dh; }
  operator SO3() const { return (SO3)dh; }

  R3      s()  const {return cog;}
  R3      ps() const {return dh.ps();}
  double  m()  const {return mass;}
  Inertia I()  const {return inertia;}
  
  SE3 theta (double t) {dh.t(t); return (SE3)dh;}
};
#endif
