#include <iostream>
#include <iomanip>
#include "globals.h"

#ifndef __MOTOR_HH__
#define __MOTOR_HH__

using namespace std;

class Motor{
private:
  pthread_mutex_t mutex;
  
  int ID;
  double q;
  double t;
  double puckI_per_Nm;       // Newton Meters -> puck torque units
  
  void lock(){pthread_mutex_lock(&mutex);};
  void unlock(){pthread_mutex_unlock(&mutex);};

 public:
  
  static const int M1 = 1;
  static const int M2 = 2;
  static const int M3 = 3;
  static const int M4 = 4;
  static const int M5 = 5;
  static const int M6 = 6;
  static const int M7 = 7;
#ifdef WRIST
  static const int Mn = 7;
#else
  static const int Mn = 4;
#endif
  
  static const float MIN_TRQ[];
  static const float MAX_TRQ[];
 
  Motor(){
    pthread_mutex_init(&mutex, NULL);
    q=0.0;   t=0.0;  ID = -1;
  }
  
  int id(){return ID;}
  double IPNm(){return puckI_per_Nm;}
  
  double trq(){double T; lock(); T = t; unlock(); return T;}
  double pos(){double Q; lock(); Q = q; unlock(); return Q;}

  void trq(double T){ 
    lock(); 
    t = T;
    //t = clip( T, Motor::MIN_TRQ[id()], Motor::MAX_TRQ[id()] ); 
    unlock();
  }
  void pos(double Q){ lock();  q = Q;  unlock(); }
  
  friend istream& operator >> (istream& s, Motor& motor){
    s >> motor.ID >> motor.puckI_per_Nm;
    return s;
  }

  friend ostream& operator << (ostream& s, Motor& motor){
    s << "id#: "  << setw(1) << motor.ID << "; " 
      << "I/Nm: " << setw(6) << motor.puckI_per_Nm;
    return s;
  }
}; 

#endif
