#include "PID.hh"

istream& PID::get(istream& s){
  s >> Kp >> Kd >> Ki >> saturation;
  return s;
}

ostream& PID::put(ostream& s){
  s << "Kp: "  << setw(4) << Kp << "; "
    << "Kd: "  << setw(4) << Kd << "; "
    << "Ki: "  << setw(4) << Ki << "; "
    << "Sat: " << saturation;
  return s;
}

PID::PID(){
  Kp = Kd = Ki = qref = laste = se = saturation = 0.0;
  firsttick = 1;
}

void PID::reset(double qref){
  lock();
  //se = 0.0;
  laste = 0.0;
  //firsttick = 1;
  this->qref = qref;
  unlock();
}

void PID::set(double qref){
  lock();
  this->qref = qref;
  unlock();
}

double PID::evaluate(double q, double dt){
  double e, de, result;
  
  lock();
  
  e = qref - q;
  if(firsttick){
    laste = e;
    firsttick = 0;
  }
  
  de = (e - laste)/dt;
  laste = e;
  se += e;
  
  if(saturation)
    if(saturation < fabs(se)*Ki)
      se = sign(se)*saturation;
  
  result = Kp*e + Kd*de + Ki*se;
  
  unlock();

  return result;
}

