#include "WAM.hh"

#define LOW_TRQ_COEFFICIENT (0.75)
static void* control_loop(void* argv);

/*
 * Create a new WAM that uses the CAN bus cb;
 */

WAM::WAM(CANbus* cb){
  rec = false;
  wsdyn = false;
  jsdyn = false;
  
  pthread_mutex_init(&mutex, NULL);

  bus = cb;
  se3traj = NULL;
  jointstraj = NULL;

  links[Link::L0]=Link( DH(  0.0000,   0.0000,   0.0000,   0.0000), 0.0000, 
			R3(  0.0000,   0.0000,   0.0000), 
			Inertia(0,0,0,0,0,0));

  links[Link::L1]=Link( DH( -M_PI_2,   0.0000,   0.0000,   0.0000), 8.3936,
 			R3(  0.3506, 132.6795,   0.6286)*0.001,
			Inertia(95157.4294,   246.1404,    -95.0183,
				92032.3524,  -962.6725,  59290.5997, M2_MM2) );

  links[Link::L2]=Link( DH(  M_PI_2,   0.0000,   0.0000,   0.0000), 4.9487,
	 		R3( -0.2230, -21.3924,  13.3754)*0.001, 
			Inertia(29326.8098,   -43.3994,   -129.2942,
				20781.5826,  1348.6924,  22807.3271, M2_MM2) );

  links[Link::L3]=Link( DH( -M_PI_2,   0.0450,   0.5500,   0.0000), 1.8251,
		 	R3(-38.7565, 217.9078,   0.0252)*0.001, 
			Inertia(56662.2970, -2321.6892,      8.2125,
			  	 3158.0509,   -16.6307,  58806.6024, M2_MM2) );
#ifdef WRIST
#warning "COMPILING WRIST"
  links[Link::L4]=Link( DH(  M_PI_2,  -0.0450,   0.0000,   0.0000), 2.1824,
			R3(  6.2895,  -0.0010, 111.0633)*0.001,
			Inertia(10065.3990,    14.6007,  -1392.4965,
				10000.5377,    29.4814,   2838.9554, M2_MM2) );

  links[Link::L5]=Link( DH( -M_PI_2,   0.0000,   0.3000,   0.0000), 0.4067,
			R3(  0.0584,  28.3754,   0.1902)*0.001, 
			Inertia(  321.0141,    -0.0667,     -0.0079,
				  172.3637,    -2.4724,    350.6782, M2_MM2) );

  links[Link::L6]=Link( DH(  M_PI_2,   0.0000,   0.0000,   0.0000), 0.5278,
			R3( -0.0311, -14.8635,  25.6326)*0.001,
			Inertia(  604.1921,     0.0825,     -0.1896,
				  269.3020,   -62.3326,    507.9036, M2_MM2) );
#else
#endif

#ifdef BH8
#warning "COMPILING HAND"
  //
  // FAT WARNING
  // I had the mass and the center of mass from old files from Barrett.
  // I figured the inertia tensor by modeling a cylinder of 13cm long and
  // 8.5cm diameter, with a uniform density computed from the mass and the
  // cylinder's volume.
  // The kinematics remains the one of J7.
  links[Link::L7]=Link( DH(  0.0000,   0.0000,   0.0620,   0.0000), 1.2000,
			R3(  0.0000,   0.0000,  45.0000)*0.001,
			Inertia(   2231.8750,   0.0000,      0.0000,
				   2231.8750,   0.0000,   1083.7500, M2_MM2) );
  /*
  links[Link::L7]=Link( DH(  0.0000,   0.0000,   0.0620,   0.0000), 1.3000,
			R3( -30.0000,   0.0000,  45.0000)*0.001,
			Inertia(   2231.8750,   0.0000,      0.0000,
				   2231.8750,   0.0000,   1083.7500, M2_MM2) );
  */
#else
  links[Link::L7]=Link( DH(  0.0000,   0.0000,   0.0620,   0.0000), 0.0557,
			R3( -0.0001,  -0.1823,-284.7431)*0.001,
			Inertia(   21.4102,    -0.0001,      0.0000,
				   21.6157,     0.0040,     42.1975, M2_MM2) );
#endif
}

/*
 * ok. the arguement for the pid*.dat files goes as follow.
 * J7 has quite a bit of friction for its weight. Thus, for small torques
 * it doesn't move at all.
 * HOWEVER! This doesn't solve the problem when moving in WS control...
 * (See the lil' hack in the control loop for this)
 */
int WAM::init(){

#ifdef WRIST

#ifdef BH8
  if(load("wam7.dat") == OW_FAILURE || loadctrl("pid7bh8.dat") == OW_FAILURE){
#else
  if(load("wam7.dat") == OW_FAILURE || loadctrl("pid7.dat") == OW_FAILURE){
#endif

#else

  if(load("wam4.dat") == OW_FAILURE || loadctrl("pid4.dat") == OW_FAILURE){

#endif
      cerr << "WAM::init: failed to load config file." << endl;
      return OW_FAILURE;

#ifdef WRIST
#ifdef BH8
  }
#else             // F%&^$$#!@  emacs...
  }               // making me do all this just to keep auto indentation
#endif
#else
  }
#endif
    return OW_SUCCESS;
}

// load the WAM file and all the joints and motor stuff...
int WAM::load(char* fn){
  ifstream is;
  
  is.open(fn);
  if(!is){
    cerr<<"WAM::load: couln't open file "<<fn<<endl;
    return OW_FAILURE;
  }
  
  is >> A >> B >> C >> D;
  
  for(int j=Joint::J1; j<=Joint::Jn; j++)
    is >> joints[j];
  
  for(int m=Motor::M1; m<=Motor::Mn; m++)
    is >> motors[m];
  
  is.close();
  
  return OW_SUCCESS;
}

int WAM::loadctrl(char* fn){
  ifstream is;
  
  is.open(fn);
  if(!is){
    cerr<<"WAM::load: couln't open file "<<fn<<endl;
    return OW_FAILURE;
  }
  for(int j=Joint::J1; j<=Joint::Jn; j++)
    is >> jointsctrl[j];
  return OW_SUCCESS;
}

/*
 * Output the information about the WAM
 */
void WAM::dump(){
  cout << "WAM information." << endl;
  cout << Joint::Jn << " joints were defined." << endl;
  for(int j=Joint::J1; j<=Joint::Jn; j++) 
    cout<<joints[j]<<endl;
  cout << "Let me tell you about the controllers I have for these." << endl;
  for(int j=Joint::J1; j<=Joint::Jn; j++) 
    cout<<jointsctrl[j]<<endl;
  cout << endl << Motor::Mn << " motors were defined." << endl;
  for(int m=Motor::M1; m<=Motor::Mn; m++) 
    cout<<motors[m]<<endl;
}

/*
 * Ask the CAN bus to query the pucks for positions and copy the result
 * in the motor positions. The positions are already converted in rads
 * (check CANBus::get_positions(double*)
 * As is, this takes about 635usec. With spikes of 4000usec. Real time my ass
 */
int WAM::recv_mpos(){
  double mpos[NUM_NODES+1]; // large enough for the all possible pucks

  // fetch positions from the bus
  if(bus->read_positions(mpos) == OW_FAILURE){
    cerr << "WAM::recv_mpos: read_positions failed." << endl;
    return OW_FAILURE;
  }

  // copy the positions into the motors' structures
  for(int m=Motor::M1; m<=Motor::Mn; m++)
    motors[m].pos( mpos[ motors[m].id() ] );

  return OW_SUCCESS;
}

/*
 * Used to send motor torques on the bus. The torques *are* converted
 * As with receiving positions, this takes quite a bit of time
 * roughly 285usec. Also spikes of 3500usec. Just brutal...
 * All in all, just sending and receiving stuff on the CAN bus takes up 1ms.
 * So forget about a 1kHz control loop...
 */
int WAM::send_mtrq(){
  long mtrq[NUM_NODES+1];

  for(int m=Motor::M1; m<=Motor::Mn; m++)
    mtrq[ motors[m].id() ] = (long) (motors[m].trq() * motors[m].IPNm());

  if(bus->send_torques(mtrq) == OW_FAILURE){
    cerr << "WAM::send_mtrq: send_torques failed." << endl;
    return OW_FAILURE;
  }

  return OW_SUCCESS;
}

// convert motors positions to joints positions
void WAM::mpos2jpos(){
  joints[1].pos( -motors[1].pos()/mN1 );
  joints[2].pos(  motors[2].pos()*0.5/mN2     -motors[3].pos()*0.5/mN3 );
  joints[3].pos( -motors[2].pos()*0.5*mn3/mN2 -motors[3].pos()*0.5*mn3/mN3 );
  joints[4].pos(  motors[4].pos()/mN4 );
#ifdef WRIST
  joints[5].pos(  motors[5].pos()*0.5/mN5     +motors[6].pos()*0.5/mN6 );
  joints[6].pos( -motors[5].pos()*0.5*mn6/mN5 +motors[6].pos()*0.5*mn6/mN6 );
  joints[7].pos( -motors[7].pos()/mN7 );
#endif
}

// convert joints positions to motors positions
void WAM::jpos2mpos(){
  motors[1].pos( -joints[1].pos()*mN1 );
  motors[2].pos(  joints[2].pos()*mN2 - joints[3].pos()*mN3/mn3 );
  motors[3].pos( -joints[2].pos()*mN2 - joints[3].pos()*mN3/mn3 );
  motors[4].pos(  joints[4].pos()*mN4 );
#ifdef WRIST
  motors[5].pos(  joints[5].pos()*mN5 - joints[6].pos()*mN6/mn6 );
  motors[6].pos(  joints[5].pos()*mN5 + joints[6].pos()*mN6/mn6 );
  motors[7].pos( -joints[7].pos()*mN7 );
#endif
}

// convert joints torques to motor torques
void WAM::jtrq2mtrq(){
  motors[1].trq( -joints[1].trq()/mN1 );
  motors[2].trq(  joints[2].trq()*0.5/mN2 - joints[3].trq()*0.5*mn3/mN3 );
  motors[3].trq( -joints[2].trq()*0.5/mN2 - joints[3].trq()*0.5*mn3/mN3 );
  motors[4].trq(  joints[4].trq()/mN4 );
#ifdef WRIST
  motors[5].trq(  joints[5].trq()*0.5/mN5 - joints[6].trq()*0.5*mn6/mN6 );
  motors[6].trq(  joints[5].trq()*0.5/mN5 + joints[6].trq()*0.5*mn6/mN6 );
  motors[7].trq( -joints[7].trq()/mN7 );
#endif
}

/*
 * Used to set/initialize the pucks positions.
 * WARNING
 * Should really backup the joints and motor values in case that 
 * CANBus::set_positions(double*) fails so that we can write them back...
 */
int WAM::set_jpos(double* pos){
  double mpos[NUM_NODES+1];

  // block the control loop from messing up with the motors and joints
  lock();
  
  // convert the joint positions to motor positions
  for(int j=Joint::J1; j<=Joint::Jn; j++)
    joints[j].pos( pos[ joints[j].id() ] );
  jpos2mpos();

  // copy the positions in the array
  for(int m=Motor::M1; m<=Motor::Mn; m++)
    mpos[ motors[m].id() ] = motors[m].pos();

  // send the position array on the CAN bus
  if(bus->send_positions(mpos) == OW_FAILURE){
    cerr << "WAM::set_jpos: bus.send_position failed." << endl;
    unlock();
    return OW_FAILURE;
  }

  unlock();
  return OW_SUCCESS;
}

void WAM::get_jpos(double* pos){
  for(int j=Joint::J1; j<=Joint::Jn; j++)
    pos[ joints[j].id() ] = joints[j].pos();
}
void WAM::get_jtrq(double* trq){
  for(int j=Joint::J1; j<=Joint::Jn; j++)
    trq[ joints[j].id() ] = joints[j].trq();
}

/*
 * Start the control timer and loop
 */
int WAM::start(){
  // start the control loop and uses control_loop (below) as the main function
  // pass this WAM as an argument to control_loop
  if(ctrl_loop.start(&control_loop, this) == OW_FAILURE){
    cerr << "WAM::start: starting control loop failed." << endl;
    if(bus->stop() == OW_FAILURE)
      cerr << "WAM::start: bus.stop() failed." << endl;
    return OW_FAILURE;
  }
  return OW_SUCCESS;
}

void WAM::stop(){
  if(ctrl_loop.stop() == OW_FAILURE)
    cerr << "WAM::stop: ctrl_loop.stop() failed." << endl;
}

/*
 * Function of the control loop thread.
 * Doesn't do much. It wakes up on the semaphore posted by the control loop
 * timer and call the WAM's control method.
 * The parameter argv is the WAM object
 */
void* control_loop(void* argv){
  WAM* wam = (WAM*)argv;                      // read the WAM argument
  ControlLoop* ctrl_loop = &(wam->ctrl_loop); // get the ControlLoop member
  RTIME t1=0, t2=0;

  // Use __LATENCY__ to record stuff for control loop latencies
#ifdef __LATENCY__
#define ARRAY_SIZE 20000
  double time[ARRAY_SIZE];
  int i=0;
#endif

  // Use __IDENTIFICATION__ to record stuff for dynamic identification
#ifdef __IDENTIFICATION__
#define ARRAY_SIZE 30000
  double time[ARRAY_SIZE];
  double positions[ARRAY_SIZE][8];
  double torques[ARRAY_SIZE][8];
  int i=0;
#endif

  // Main control loop

  while(ctrl_loop->state() == CONTROLLOOP_RUN){
    ctrl_loop->wait();         // wait for the next control iteration;
    t2 = rt_get_cpu_time_ns(); // record the time
    
#ifdef __LATENCY__
    if(wam->record())    time[i++] = ((double)t2)/1000000000.0;
    if(i == ARRAY_SIZE){
      cout << "WAM::control_loop: recording done." << endl;
      wam->record() = false;
      i++;
    }
#endif
#ifdef __IDENTIFICATION__
    if(wam->record()){
      time[i] = ((double)t2)/1000000000.0;
      wam->get_jpos(&positions[i][0]);
    }
#endif

    // This is where the control stuff happens
    wam->control( ((double)(t2-t1))/1000000000.0 ); 

#ifdef __IDENTIFICATION__
    if(wam->record()){
      wam->get_jtrq(&torques[i][0]);
      i++;
    }
    if(i == ARRAY_SIZE){
      cout << "WAM::control_loop: recording done." << endl;
      wam->record() = false;
      i++;
    }
#endif

    t1 = t2;
  }

#ifdef __LATENCY__
  FILE* fp = fopen("latencies_dir/latencies", "w");
  fwrite((void*)time, sizeof(double), ARRAY_SIZE, fp);
  fclose(fp);
  cout << "Data dumped in file: latencies_dir/latencies." << endl;
#endif
#ifdef __IDENTIFICATION__
  FILE* fp = fopen("identification", "w");
  fwrite((void*)time,             sizeof(double), ARRAY_SIZE,   fp);
  fwrite((void*)&positions[0][0], sizeof(double), ARRAY_SIZE*8, fp);
  fwrite((void*)&torques[0][0],   sizeof(double), ARRAY_SIZE*8, fp);
  fclose(fp);
  cout << "Data dumped in file: identification" << endl;
#endif

  return NULL;
}


/*
 * This is the function called at each control loop
 * All the control stuff should happen in here
 */
void WAM::control(double dt){
  double   t[Joint::Jn+1];
  double   q[Joint::Jn+1];
  double  qd[Joint::Jn+1];
  double qdd[Joint::Jn+1];
  R6 F;

  lock();

  if(recv_mpos() == OW_SUCCESS){            // read new motor positions
  
    for(int j=Joint::J1; j<=Joint::Jn; j++)
      q[j]  = joints[j].pos();              // save the old joints positions
    
    mpos2jpos();                            // convert new joint positions

    for(int j=Joint::J1; j<=Joint::Jn; j++){
      qd[j]  = (joints[j].pos() - q[j])/dt; // compute the joint velocities
      q[j]   =  joints[j].pos();            // read the new joints positions
      qdd[j] = 0.0;
    }

    for(int l=Link::L1; l<=Link::Ln; l++)   // compute the FK
      links[l].theta( q[l] );
    E0n = forward_kinematics(links);

    // is there a Cartesian trajectory running
    if(se3traj != NULL){
      F = WSControl(dt);
      if(se3traj->state() == Trajectory::STOP){ // check if the trajectory is 
	delete se3traj;                         // over. Might be a good thing
	se3traj = NULL;                         // to have a "sticky" flag that
	se3ctrl.stop();                         // determines if the controller
                                                // should keep running
	// this is a useful message (shouldn't be in the control loop though)
	cout << "Controllers are off" << endl;
      }
    }
    // is there a joint trajectory running
    else if(jointstraj != NULL){
      JSControl(qdd, dt);
      if(jointstraj->state() == Trajectory::STOP){
	delete jointstraj;
	jointstraj = NULL;
	for(int j=Joint::J1; j<=Joint::Jn; j++)  jointsctrl[j].stop();
	// this is a useful message (shouldn't be in the control loop though)
	cout << "Controllers are off" << endl;
      }
    }
    
    if(wsdyn){                               // compute WS dynamics
      WSdynamics(t, links, qd, F);
      for(int j=Joint::J1; j<=Joint::Jn; j++)
	joints[j].trq(t[j]);                 // write the torques in the joints

      if(se3ctrl.state() == Controller::RUN){// J7 has friction so I add this
        double t7 = t[Joint::J7]*100;        // to make up for it (works better
        joints[Joint::J7].trq(t7);           // in practice)
      }
    }

    else if(jsdyn){                          // compute JS dynamics
      JSdynamics(t, links, qd, qdd); 
      for(int j=Joint::J1; j<=Joint::Jn; j++){
	joints[j].trq(t[j]);                 // write the torques in the joints
      }
    }

    jtrq2mtrq();                             // results in motor::torque

    if(send_mtrq() == OW_FAILURE)
      cerr << "WAM::control_loop: send_mtrq failed." << endl;
  }
  
  unlock();
}
 
R6 WAM::WSControl(double dt){
  R3 pd, wd, pdd, wdd;  // desired velocities and accelerations
  SE3 E0ns = E0n;       // desired forward kinematics (initialized to FK
                        // in case the trajectory is not running but the 
                        // controller below is)
  
  // if the trajectory isn't running all values are left intact
  se3traj->evaluate(E0ns, pd, pdd, wd, wdd, dt);
  SO3 Rn0 = !((SO3)E0n);
  pdd = Rn0*pdd;        // align the linear acc in the body frame
  wdd = Rn0*wdd;        // align the angular acc in the body frame
  
  R6 F( pdd, wdd );                     // feedforward forces/moments

  // if the controller isn't running the values are left intact
  F = F + se3ctrl.evaluate(E0ns, E0n, dt); // + feedback forces/moments

  return F;
}

void WAM::JSControl(double qdd[Joint::Jn+1], double dt){
  double q[Joint::Jn+1], qd[Joint::Jn+1];

  for(int j=Joint::J1; j<=Joint::Jn; j++){
    qdd[j] = qd[j] = 0.0;
    q[j] = joints[j].pos();   // initialize the position to the current values
  }                           // in case that no trajectory is running and the
                              // controllers are
  // if no trajectory is running, then all values are left intact
  jointstraj->evaluate(q, qd, qdd, dt);     // feedforward acceleration

  // if the controllers aren't running the values are left intact
  for(int j=Joint::J1; j<=Joint::Jn; j++)   // + feedback
    qdd[j] = qdd[j] + jointsctrl[j].evaluate( q[j], joints[j].pos(), dt );
}

/*
 * Move methods. They don't do much except to allocate trajectories and
 * their respective profiles.
 * Whatever you do. Make sure you don't leave any these allocated objects
 * within reach of the user and let the control loop handle deleting them. 
 * The trajectories will be deleted when their done, and the destructors will
 * handle deleting the profiles.
 * It's safer this way...trust me...I know...
 */

void WAM::move_sigmoid(const SE3& E02){
  se3traj = new SE3Traj( FK(), E02, 
			 new Sigmoid(Sigmoid::VMAX), 
			 new Sigmoid(Sigmoid::WMAX) );
  se3traj->run();
  se3ctrl.reset();
  se3ctrl.set(FK());
  se3ctrl.run();
}

void WAM::move_sigmoid(const double q2[Joint::Jn+1]){
  Profile* profs[Joint::Jn+1];
  double q1[Joint::Jn+1];

  for(int j=Joint::J1; j<=Joint::Jn; j++){
    q1[j] = joints[j].pos();
    profs[j] = new Sigmoid(0.1);
  }
  
  jointstraj = new JointsTraj(q1, q2, profs);
  jointstraj->run();

  for(int j=Joint::J1; j<=Joint::Jn; j++){
    jointsctrl[j].reset();
    jointsctrl[j].set(joints[j].pos());
    jointsctrl[j].run();
  }
}

/*
void WAM::moveTrapezoid(const SE3& E02){
  se3traj = new SE3Traj( FK(), E02, 
			 new Trapezoid(Trapezoid::VMAX), 
			 new Trapezoid(Trapezoid::WMAX) );
  se3traj->run();
  se3ctrl.reset();
  se3ctrl.set(FK());
  se3ctrl.run();
}5A
void WAM::moveTrapezoid(const double* q2[Joint::Jn+1]){
  for(int j=Joint::J1; j<=Joint::Jn; j++){
    if(mask & JointTraj::MASK[j]){
      jointtraj[j] = new JointTraj(joints[j].pos(), q2[j], 
				   new Trapezoid(0.1) );
      jointtraj[j]->run();
      jointctrl[j].reset();
      jointctrl[j].set(joints[j].pos());
      jointctrl[j].run();
    }
    else jointtraj[j] = NULL;
  }
}
*/

void WAM::printjpos(){
  cout.setf(ios::fixed);
  cout.precision(6);
  for(int j=Joint::J1; j<=Joint::Jn; j++)
    cout << "J" << j << ": " << joints[j].pos() << " ";
  cout<<endl;
}
void WAM::printmtrq(){
  cout.setf(ios::fixed);
  cout.precision(2);
  for(int m=Motor::M1; m<=Motor::Mn; m++)
    cout << "M" << m << ": " << motors[m].trq() << " ";
  cout<<endl;
}

