#include <iostream>
#include <stdio.h>
#include <semaphore.h>
#include <list>

#include "CANbus.hh"
#include "Joint.hh"
#include "Motor.hh"
#include "ControlLoop.hh"
#include "Link.hh"
#include "Kinematics.hh"
#include "Dynamics.hh"

#include "SE3Traj.hh"
#include "SE3CtrlPD.hh"
#include "JointsTraj.hh"
#include "JointCtrlPID.hh"

#include "Sigmoid.hh"

#ifndef __WAM_H__
#define __WAM_H__

using namespace std;

class WAM{
private:
  static const double mN1 = 42.00;  //joint ratios
  static const double mN2 = 28.25;
  static const double mN3 = 28.25;
  static const double mN4 = 18.00;
  static const double mN5 = 10.27;
  static const double mN6 = 10.27;
  static const double mN7 = 14.93;
  static const double mn3 =  1.68;
  static const double mn6 =  1.00;

  static const double M2_MM2 = 1.0/1000000;

  CANbus* bus;                             // pointer to the CAN bus
   
  pthread_mutex_t mutex;                   // Use to lock the WAM
  Joint joints[Joint::Jn+1];               // Array of joints
  Motor motors[Motor::Mn+1];               // Array of motors
  Link links[Link::Ln+1];                  // Array of links
   
  double A,B,C,D,E,F;                      // WTF?
  SE3 E0n;                                 // forward kinematics transformation

  SE3Traj     *se3traj;                    // Cartesian trajectory
  SE3CtrlPD    se3ctrl;                    // basic Cartesian controller
  JointsTraj  *jointstraj;                 // joints trajectories
  JointCtrlPID jointsctrl[Joint::Jn+1];    // joint controllers

  bool rec;                                // Are we recording data
  bool wsdyn;                              // Is the WS dynamics turned on?
  bool jsdyn;                              // Is the JS dynamics turned on?

  int load(char* fn);
  int loadctrl(char* fn);
   
  int recv_mpos();
  int send_mtrq();

  void mpos2jpos();
  void jpos2mpos();   
  void jtrq2mtrq();

  R6 WSControl(double dt);
  void JSControl(double qdd[Joint::Jn+1], double dt);

  void lock(){  pthread_mutex_lock(&mutex); }
  void unlock(){  pthread_mutex_unlock(&mutex); }

public:
  ControlLoop ctrl_loop;            // control loop object

  WAM(CANbus* cb);
   
  int init();                       // initialise the WAM
  void dump();                      // print information of the WAM

  void printjpos();                 // print joints positions
  void printmtrq();                 // print motor torques

  int start();                      // start the control loop
  void stop();                      // stop the control loop
  void control(double dt);          // main control function

  int  set_jpos(double pos[Joint::Jn+1]);   // set the joints positions
  void get_jpos(double pos[Joint::Jn+1]);   // get the joints positions
  void get_jtrq(double trq[Joint::Jn+1]);   // get the joints torques

  void move_sigmoid(const SE3& E0ns);
  void move_sigmoid(const double q[Joint::Jn+1]);
  void move_trapezoid(const SE3& E0ns);
  void move_trapezoid(const double q[Joint::Jn+1]);

  bool& wsdynamics(){return wsdyn;} // use to set the dynamics
  bool& jsdynamics(){return jsdyn;} // use to set the dynamics
  
  bool& record(){return rec;}       // use to set the recording
  SE3 FK(){  SE3 e0n; lock();   e0n=E0n;   unlock();   return e0n;  }

  friend ostream& operator << (ostream& s, WAM& wam);
};




































/* Define a structure to hold the output of the whereAmI() routine */
/*
  typedef struct { 
  double x;                // Translation in x direction
  double y;                // Translation in y direction
  double z;                // Translation in z direction
  double rotMatrix[3][3];  // The rotation matrix
  }frame; 
*/

/* 
 * Define WAM types for the whereAmI() routine.
 * 4DOF    = 4-DOF WAM with standard outer link
 * 4DOF_G  = 4-DOF WAM with gimbals on outer link
 * 7DOF    = 7-DOF WAM
 */
/*
  enum {WAM_4DOF, WAM_4DOF_G, WAM_7DOF} wamType;

  int wam_init(WAM* wam);
  void wam_close(WAM* wam);
  int wam_load(WAM* wam, char *fn);
  int wam_save(WAM* wam, char* fn);
  void wam_dump(WAM* wam);

  void wam_control_loop(void *argv);

  int wam_trqeff_register(WAM* wam, torque_effector TE);
  void wam_trqeff_deregister(WAM* wam, int index);
  void wam_trqeff_process(WAM* wam);

  int wam_set_limits(WAM* wam, double jointVel, double tipVel, double elbowVel);
*/ 


//Calculate gravity compensation torques
/*
  void wam_Gcomp(WAM* wam, double A, double B, double C, double D);
  void wam_Gcomp2(WAM* wam, double* t2, double* t3, double* t4);
  int wam_Gcomp_sample(WAM* wam, double* res,
  double p1, double p2, double p3, double p4);

  void wam_setG(WAM* wam, double A,double B,double C,double D);
  void wam_getG(WAM* wam, double* A, double* B, double* C, double* D);

  void wam_Gcollect(WAM* wam, const char *fn, int j2, int j3, int j4);

  int wam_findG(WAM* wam);
  void getOcompValues(double *E,double *F);
  void wam_toggle_Ocomp(WAM* wam);

  // modes
  int wam_power(WAM* wam);
  int wam_pid(WAM* wam);
  int wam_torque(WAM* wam);
  int wam_idle(WAM* wam);

  int wam_zero(WAM* wam);
  int wam_park(WAM* wam);
  int wam_move(WAM* wam, double* pos, double* vel, double* acc);

  void wam_get_joints(WAM* wam, 
  double* Mpos, double* Mtrq,
  double* Jpos, double* Jtrq);

  void wam_get_Jpos(WAM* wam, double* Jpos);
  void wam_get_Mtrq(WAM* wam, double* Mtrq);
  void wam_get_Jtrq(WAM* wam, double* Jtrq);
  int wam_set_pos(WAM* wam, double* pos);

  void FindWAMstop(int idx);
  void FindWAMcenter(int idx);
  void GentleFindWAMcenter(int idx);
  void GentleFindWAMstop(int idx);

  int whereAmI(frame *theFrame, int wamType);
*/

#endif

